"""
Further subtyping of `GenericSurface` to be either an `IrregularSurface`
or `RegularSurface` depending on mesh connectivity.
"""

from dataclasses import dataclass
from os import PathLike
from typing import TypeVar, Generic, Sequence, Optional

from civet.abstract_data import AbstractDataCommand
from civet.bases import DataSource
from civet.extraction.side import Side
from civet.obj import GenericSurface

_IS = TypeVar('_IS', bound='GenericIrregularSurface')
_RS = TypeVar('_RS', bound='GenericRegularSurface')


@dataclass(frozen=True)
class Tetra(DataSource):

    center: tuple[float, float, float] = (0.0, 0.0, 0.0)
    radius: tuple[float, float, float] = (1.0, 1.0, 1.0)
    triangles: int = 81920

    def command(self, output: str | PathLike) -> Sequence[str | PathLike]:
        return 'create_tetra', output, *self.center, *self.radius


class RegularSurface(GenericSurface[_RS], Generic[_RS]):
    """
    Represents a mesh (`.obj`) with standard connectivity.

    Typically, standard connectivity means 81,920 triangles, 41,962
    vertices. By convention, the file name for such a surface should
    have the suffix `_81920.obj`.

    A general definition for "standard connectivity" would be a
    polygonal mesh of *N* triangles where 320 and 4 are common
    denominators of *N*.
    """
    @classmethod
    def create_tetra(cls, tetra: Tetra) -> 'RegularSurface[RegularSurface]':
        return cls(tetra)


class IrregularSurface(GenericSurface[_IS], Generic[_IS]):
    """
    Represents a mesh (`.obj`) with irregular connectivity.
    """
    def interpolate_with_sphere(self, side: Optional[Side] = None) -> RegularSurface:
        """
        Resample this surface to have a standard number of 81,920 triangles.
        """
        class InterpolatedFromSphere(RegularSurface):
            def command(self, output: str | PathLike
                        ) -> Sequence[str | PathLike | AbstractDataCommand]:
                side_option = []
                if side is not None:
                    side_option = ['-' + side.value]
                return 'interpolate_surface_with_sphere.pl', *side_option, self.input, output
        return InterpolatedFromSphere(self)



"""
Manual control of memoization features.
"""

from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import ContextManager, Sequence, Callable, NewType

from civet.abstract_data import AbstractDataCommand
from civet.shells import Shell, subprocess_run

_IntermediatePath = NewType('IntermediatePath', Path)


@dataclass(frozen=True)
class Memoizer:
    """
    A `Memoizer` executes `civet.abstract_data.AbstractDataCommand` and
    writes their outputs to temporary paths. These outputs are cached,
    so that when given the same `civet.abstract_data.AbstractDataCommand`
    to run again, its cached output is returned.

    ### Dependency Tree

    `civet.abstract_data.AbstractDataCommand.command` produces a sequence
    which usually contain other `AbstractDataCommand`.
    These nested objects are dependencies which need to be computed first.
    Hence, `AbstractDataCommand` can be thought of as
    nodes of a *dependency tree* where the root is the desired output and the
    leaves are input files. `Memoizer` performs DFS on the tree, executing
    the commands represented by each node, to produce the intermediate outputs
    necessary to compute the root.
    """

    temp_dir: Path
    shell: Shell
    require_output: bool = True
    _cache: dict[AbstractDataCommand, _IntermediatePath] = field(init=False, default_factory=dict)

    def save(self, d: AbstractDataCommand, output: str | PathLike) -> None:
        """
        If `d` was previously computed, copy the cached result to `output`.
        Else, compute `d`, cache the result, and copy to `output`.
        """
        if d not in self._cache:
            self._force_save(d)
        self.shell(('cp', '-r', self._cache[d], output))

    def _force_save(self, d: AbstractDataCommand) -> _IntermediatePath:
        """
        Compute `d` and cache the result.
        """
        output = self.__temp(d.preferred_suffix)
        cmd = self._resolve_command(d.command(output))
        self.shell(cmd)
        if self.require_output and not output.exists():
            print(f'output is: {output}')
            self.shell(('ls', self.temp_dir))
            raise NoOutputError(d)
        self._cache[d] = output
        return output

    def _cache_hit(self, d: AbstractDataCommand) -> _IntermediatePath:
        """
        If `d` was previously computed, return the path to its cached result.
        Else, compute `d` first and then return the path to its cached result.
        """
        if d in self._cache:
            return self._cache[d]
        return self._force_save(d)

    def _resolve_command(self, cmd: Sequence[str | PathLike | AbstractDataCommand]) -> Sequence[str | PathLike]:
        """
        Replace every `AbstractDataCommand` in `cmd` with a path to their cached output.
        The `AbstractDataCommand` will be computed if it was not computed before.
        """
        return tuple(self._resolve_component(c) for c in cmd)

    def _resolve_component(self, c: str | PathLike | AbstractDataCommand) -> str | PathLike:
        if isinstance(c, (str, PathLike)):
            return c
        elif isinstance(c, AbstractDataCommand):
            return self._cache_hit(c)
        # TODO subshell support
        raise TypeError(f'{c} is not a [str | PathLike | AbstractDataCommand]')

    def __temp(self, suffix='') -> _IntermediatePath:
        """
        Create a temporary path name.
        """
        with NamedTemporaryFile(suffix=suffix, dir=self.temp_dir) as t:
            pass
        return _IntermediatePath(Path(t.name))


@dataclass(frozen=True)
class Session(ContextManager[Memoizer]):
    """
    A `Session` manages the temporary directory of a `Memoizer`.
    """

    require_output: bool = True
    """
    If True, raise `NoOutputError` if a command fails to produce output to its given path.
    """
    shell: Shell = subprocess_run
    """
    A function which executes its parameters as a subprocess.
    """
    temp_dir: ContextManager[str] = field(default_factory=TemporaryDirectory)

    def __enter__(self) -> Memoizer:
        temp_dir_name = self.temp_dir.__enter__()
        return Memoizer(Path(temp_dir_name), require_output=self.require_output, shell=self.shell)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.temp_dir.__exit__(exc_type, exc_val, exc_tb)


class NoOutputError(Exception):
    """
    Raised when a subprocesses ran by `Memoizer` does not create its output given path.
    """
    pass



"""
Object-oriented bindings for
[CIVET](http://www.bic.mni.mcgill.ca/ServicesSoftware/CIVET).

See [https://pypi.org/project/pycivet/](https://pypi.org/project/pycivet/)
for a description and examples.
"""

import civet.extraction.starting_models as starting_models

__docformat__ = 'numpy'

__all__ = [
    'extraction',
    'minc',
    'obj',
    'xfm',
    'starting_models',
    'globals',
    'bases',
    'memoization',
    'abstract_data',
    'memoization',
    'shells'
]
